# Fairyfly: A Plugin for Environmental Analysis (GPL)
# This file is part of Fairyfly.
#
# Copyright (c) 2025, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Fairyfly; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>


"""
Dump any fairyfly object to a JSON file. You can use "FF Load Objects" component
to load the objects from the file back into Grasshopper.
-
Fairyfly objects include any fairyfly Model, Material, or Boundary.
-

    Args:
        _ff_objs: A list of Fairyfly objects to be written to a file.
        _name_: A name for the file to which the fairyfly objects will be
            written. (Default: 'unnamed').
        _folder_: An optional directory into which the fairyfly objects will be
            written.  The default is set to the default simulation folder.
        indent_: An optional positive integer to set the indentation used in the
            resulting JSON file.
        abridged_: Set to "True" to serialize the object in its abridged form.
            Abridged objects cannot be reserialized back to fairyfly objects
            on their own but they are used throughout fairyfly to minimize
            file size and unnecessary duplication.
        _dump: Set to "True" to save the fairyfly objects to file.

    Returns:
        report: Errors, warnings, etc.
        ff_file: The location of the file where the fairyfly JSON is saved.
"""

ghenv.Component.Name = 'FF Dump Objects'
ghenv.Component.NickName = 'DumpObjects'
ghenv.Component.Message = '1.9.0'
ghenv.Component.Category = 'Fairyfly'
ghenv.Component.SubCategory = '0 :: Create'
ghenv.Component.AdditionalHelpFromDocStrings = '4'

import sys
import os
import json

try:  # import the core fairyfly dependencies
    from fairyfly.model import Model
    from fairyfly.shape import Shape
    from fairyfly.boundary import Boundary
    from fairyfly.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly:\n\t{}'.format(e))

try:  # import the core ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def geo_object_warning(obj):
    """Give a warning that individual geometry objects should be added to a Model."""
    msg = 'An individual {} has been connected to the _ff_objs.\n' \
        'The recommended practice is to add this object to a Model and\n' \
        'serialize the Model instead of serializing individual objects.'.format(
            obj.__class__.__name__)
    print(msg)
    give_warning(ghenv.Component, msg)


if all_required_inputs(ghenv.Component) and _dump:
    # set the component defaults
    name = _name_ if _name_ is not None else 'unnamed'
    file_name = '{}.json'.format(name) if len(_ff_objs) > 1 or not \
        isinstance(_ff_objs[0], Model) else '{}.ffjson'.format(name)
    folder = _folder_ if _folder_ is not None else folders.default_simulation_folder
    ff_file = os.path.join(folder, file_name)
    abridged = bool(abridged_)

    # check to see if any objects are of the geometry type and give a warning
    geo_types = (Shape, Boundary)
    for obj in _ff_objs:
        if isinstance(obj, geo_types):
            geo_object_warning(obj)

    # create the dictionary to be written to a JSON file
    if len(_ff_objs) == 1:  # write a single object into a file if the length is 1
        try:
            obj_dict = _ff_objs[0].to_dict(abridged=abridged)
        except TypeError:  # no abridged option
            obj_dict = _ff_objs[0].to_dict()
    else:  # create a dictionary of the objects that are indexed by name
        obj_dict = {}
        for obj in _ff_objs:
            try:
                obj_dict[obj.identifier] = obj.to_dict(abridged=abridged)
            except TypeError:  # no abridged option
                obj_dict[obj.identifier] = obj.to_dict()

    # write the dictionary into a file
    if not os.path.isdir(folder):
        os.makedirs(folder)
    if (sys.version_info < (3, 0)):  # we need to manually encode it as UTF-8
        with open(ff_file, 'wb') as fp:
            obj_str = json.dumps(obj_dict, indent=indent_, ensure_ascii=False)
            fp.write(obj_str.encode('utf-8'))
    else:
        with open(ff_file, 'w', encoding='utf-8') as fp:
            obj_str = json.dump(obj_dict, fp, indent=indent_, ensure_ascii=False)



# Fairyfly: A Plugin for Environmental Analysis (GPL)
# This file is part of Fairyfly.
#
# Copyright (c) 2025, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Fairyfly; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>


"""
Load any fairyfly object from a fairyfly JSON file
-
Fairyfly objects include any Model, Shape, or Boundary.
-
It also includes any fairyfly Material or Simulation object.
-

    Args:
        _ff_file: A file path to a fairyfly JSON from which objects will be loaded
            back into Grasshopper. The objects in the file must be non-abridged
            in order to be loaded back correctly.
        _load: Set to "True" to load the objects from the _ff_file.

    Returns:
        report: Reports, errors, warnings, etc.
        ff_objs: A list of fairyfly objects that have been re-serialized from
            the input file.
"""

ghenv.Component.Name = 'FF Load Objects'
ghenv.Component.NickName = 'LoadObjects'
ghenv.Component.Message = '1.9.1'
ghenv.Component.Category = 'Fairyfly'
ghenv.Component.SubCategory = '0 :: Create'
ghenv.Component.AdditionalHelpFromDocStrings = '4'

import io

try:  # import the core fairyfly dependencies
    import fairyfly.dictutil as ff_dict_util
    from fairyfly.model import Model
    from fairyfly.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly:\n\t{}'.format(e))

try:  # import the core fairyfly_energy dependencies
    import fairyfly_therm.material.dictutil as therm_dict_util
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly_energy:\n\t{}'.format(e))

try:  # import the core ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
    from ladybug_rhino.config import units_system, current_tolerance
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

import json


def model_units_tolerance_check(model):
    """Convert a model to the current Rhino units and check the tolerance.

    Args:
        model: A fairyfly Model, which will have its units checked.
    """
    # check the model units
    if model.units != units_system():
        print('Imported model units "{}" do not match that of the current Rhino '
            'model units "{}"\nThe model is being automatically converted '
            'to the Rhino doc units.'.format(model.units, units_system()))
        model.convert_to_units(units_system())

    # check that the model tolerance is not too far from the Rhino tolerance
    if model.tolerance / current_tolerance() >= 100:
        msg = 'Imported Model tolerance "{}" is significantly coarser than the ' \
            'current Rhino model tolerance "{}".\nIt is recommended that the ' \
            'Rhino document tolerance be changed to be coarser and this ' \
            'component is re-run.'.format(model.tolerance, current_tolerance())
        print msg
        give_warning(ghenv.Component, msg)


if all_required_inputs(ghenv.Component) and _load:
    with io.open(_ff_file, encoding='utf-8') as inf:
        first_char = inf.read(1)
        second_char = inf.read(1)
    with io.open(_ff_file, encoding='utf-8') as inf:
        if second_char == '{':
            inf.read(1)
        data = json.load(inf)

    if 'type' in data:
        ff_objs = ff_dict_util.dict_to_object(data, False)  # re-serialize as a core object
        if ff_objs is None:  # try to re-serialize it as an energy object
            ff_objs = therm_dict_util.dict_to_object(data, False)
        elif isinstance(ff_objs, Model):
            model_units_tolerance_check(ff_objs)
    else:  # no 'type' key; assume that its a group of objects
        ff_objs = []
        for hb_dict in data.values():
            ff_obj = ff_dict_util.dict_to_object(hb_dict, False)  # re-serialize as a core object
            if ff_obj is None:  # try to re-serialize it as an energy object
                ff_obj = therm_dict_util.dict_to_object(hb_dict, False)
            ff_objs.append(ff_obj)


# Fairyfly: A Plugin for Environmental Analysis (GPL)
# This file is part of Fairyfly.
#
# Copyright (c) 2025, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Fairyfly; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a custom cavity material, which can be plugged into the "FF Shape" component
to change the properties of the material in a THERM simulation.
-

    Args:
        _name_: Text to set the name for the material.
        _cavity_model_: Text for the type of cavity model to be used to determine
            the thermal resistance of the material. (Default: CEN). Choose from
            the following.
                * CEN
                * NFRC
                * ISO15099
                * ISO15099Ventilated
        _gas_type_: Text describing the types of gas in the gap. (Default: Air).
            Text must be one of the following: 'Air', 'Argon', 'Krypton', 'Xenon'.
        _gas_ratios_: A list of text describing the volumetric fractions of gas
            types in the mixture.  This list must align with the gas_types
            input list. Default: Equal amout of gases for each type.
        rgb_color_: An optional color to set the color of the material when it is
            imported to THERM. If unspecified, a randomly-generated color is assigned.

    Returns:
        mat: A cavity material that can be assigned to a Fairyfly Shape to change the
            properties of the material in a THERM simulation.
"""

ghenv.Component.Name = 'FF Cavity Material'
ghenv.Component.NickName = 'CavityMat'
ghenv.Component.Message = '1.9.0'
ghenv.Component.Category = 'Fairyfly'
ghenv.Component.SubCategory = '1 :: THERM'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

import uuid

try:  # import the core ladybug dependencies
    from ladybug.color import Color
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:  # import the fairyfly-energy dependencies
    from fairyfly_therm.material import CavityMaterial, Gas
    from fairyfly_therm.lib.gases import gas_by_name
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly_energy:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import turn_off_old_tag
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))
turn_off_old_tag(ghenv.Component)


# set the default material properties
_cavity_model_ = 'CEN' if _cavity_model_ is None else _cavity_model_
_gas_type_ = 'Air' if _gas_type_ is None else _gas_type_

# get the Gas material to fill the cavity
cavity_gas = gas_by_name(_gas_type_.title())

# create the material
mat = CavityMaterial(cavity_gas, _cavity_model_)
mat.display_name = _name_ if _name_ is not None else \
    'Cavity Material {}'.format(str(uuid.uuid4())[:8])
if rgb_color_ is not None:
    mat.color = Color(rgb_color_.R, rgb_color_.G, rgb_color_.B)



# Fairyfly: A Plugin for Environmental Analysis (GPL)
# This file is part of Fairyfly.
#
# Copyright (c) 2025, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Fairyfly; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Color Fairyfly Boundaries in the Rhino scene using their attributes.
_
This can be used as a means to check that correct properties are assigned to
different boundaries.
-

    Args:
        _ff_objs: An array of fairyfly Boundaries to be colored with their attributes
            in the Rhino scene. This can also be an entire Model to be colored.
        _attribute_: Text for the name of the attribute with which the boundaries should
            be colored. The "FF Boundary Attributes" component lists all of the
            attributes of the boundary. (Default: THERM Condition).
        legend_par_: An optional LegendParameter object to change the display
            of the colored boundaries (Default: None).

    Returns:
        lines: Line segments of the boundaries colored according to their attributes.
        legend: Geometry representing the legend for colored lines.
        wire_frame: A list of lines representing the outlines of the _ff_objs.
        values: A list of values noting the attribute assigned to each boundary.
        colors: A list of colors noting the color of each boundary in the Rhino scene.
            This can be used in conjunction with the native Grasshopper
            "Custom Preview" component to create custom visualizations in
            the Rhino scene.
        vis_set: An object containing VisualizationSet arguments for drawing a detailed
            version of the ColorRoom in the Rhino scene. This can be connected to
            the "LB Preview Visualization Set" component to display this version
            of the visualization in Rhino.
"""

ghenv.Component.Name = 'FF Color Boundary Attributes'
ghenv.Component.NickName = 'ColorBoundaryAttr'
ghenv.Component.Message = '1.9.1'
ghenv.Component.Category = 'Fairyfly'
ghenv.Component.SubCategory = '0 :: Create'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:  # import the core fairyfly dependencies
    from fairyfly.model import Model
    from fairyfly.boundary import Boundary
    from fairyfly.shape import Shape
    from fairyfly.colorobj import ColorBoundary
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly:\n\t{}'.format(e))

try:  # import the ladybug_rhino dependencies
    from ladybug_rhino.colorize import ColoredLine
    from ladybug_rhino.fromgeometry import from_linesegment2d, from_face3d_to_wireframe
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.color import color_to_color
    from ladybug_rhino.grasshopper import all_required_inputs, schedule_solution
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

DEFAULT_PROP = 'properties.therm.condition.display_name'


if all_required_inputs(ghenv.Component):
    # extract any boundaries from input Models
    _attribute_ = DEFAULT_PROP if _attribute_ is None else _attribute_
    boundaries, shapes = [], []
    for ff_obj in _ff_objs:
        if isinstance(ff_obj, Model):
            boundaries.extend(ff_obj.boundaries)
            shapes.extend(ff_obj.shapes)
        elif isinstance(ff_obj, Boundary):
            boundaries.append(ff_obj)
        elif isinstance(ff_obj, Shape):
            shapes.append(ff_obj)

    # create the ColorBoundary visualization object
    color_obj = ColorBoundary(boundaries, _attribute_, legend_par_)
    # assign condition colors if this is what is being used for coloring
    if _attribute_ == DEFAULT_PROP and color_obj.legend_parameters.are_colors_default:
        color_model = Model(boundaries=boundaries)
        color_dict = {con.display_name: con.color
                      for con in color_model.properties.therm.conditions}
        leg_colors = [color_dict[m_name] for m_name in color_obj.attributes_unique]
        if len(leg_colors) == 1:
            leg_colors = [leg_colors[0]] * 2
        color_obj.legend_parameters.colors = leg_colors

    # output the visualization geometry
    graphic = color_obj.graphic_container
    values = color_obj.attributes_original
    flat_geo = color_obj.flat_geometry
    wire_frame = []
    for shape in shapes:
        wire_frame.extend(from_face3d_to_wireframe(shape.geometry))
    legend = legend_objects(graphic.legend)
    colors = [color_to_color(col) for col in graphic.value_colors]
    lines = []
    for lin, col in zip(flat_geo, colors):
        col_line = ColoredLine(from_linesegment2d(lin))
        col_line.color = col
        col_line.thickness = 3
        lines.append(col_line)
    # CWM: I don't know why we have to re-schedule the solution but this is the
    # only way I found to get the colored polylines to appear (redraw did not work).
    schedule_solution(ghenv.Component, 2)
    vis_set = color_obj



# Fairyfly: A Plugin for Environmental Analysis (GPL)
# This file is part of Fairyfly.
#
# Copyright (c) 2025, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Fairyfly; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>


"""
Deconstruct a material into its constituient attributes and values.
-

    Args:
        _mat: A material to be deconstructed. This can also be text for a
            material to be looked up in the material library.

    Returns:
        values: List of values for the attributes that define the material.
        attr_names: List of text that is the same length as the values, which
            notes the attribute name for each value.
"""

ghenv.Component.Name = 'FF Deconstruct Material'
ghenv.Component.NickName = 'DecnstrMat'
ghenv.Component.Message = '1.9.0'
ghenv.Component.Category = 'Fairyfly'
ghenv.Component.SubCategory = '1 :: THERM'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

import re

try:  # import the fairyfly-therm dependencies
    from fairyfly_therm.lib.materials import solid_material_by_name, cavity_material_by_name
    from fairyfly_therm.material import SolidMaterial
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly_therm:\n\t{}'.format(e))

try:  # import the fairyfly dependencies
    from fairyfly.search import get_attr_nested
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

SOLID_ATTR = (
    'display_name',
    'conductivity',
    'emissivity',
    'density',
    'porosity',
    'specific_heat',
    'vapor_diffusion_resistance'
)

CAVITY_ATTR = (
    'display_name',
    'cavity_model',
    'gas.pure_gases',
    'gas.gas_fractions'
)


if all_required_inputs(ghenv.Component):
    # check the input
    if isinstance(_mat, str):
        try:
            _mat = solid_material_by_name(_mat)
        except ValueError:
            _mat = cavity_material_by_name(_mat)

    # get the values and attribute names
    rel_attr = SOLID_ATTR if isinstance(_mat, SolidMaterial) else CAVITY_ATTR
    values, attr_names = [], []
    for r_attr in rel_attr:
        values.append(get_attr_nested(_mat, r_attr))
        attr_names.append(r_attr.split('.')[-1].replace('_', ' ').title())




# Fairyfly: A Plugin for Environmental Analysis (GPL)
# This file is part of Fairyfly.
#
# Copyright (c) 2025, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Fairyfly; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Write a fairyfly Model to an THMZ file (aka. THERM Zip file), which can then be
simulated or opened in the THERM interface.

-
    Args:
        _model: A fairyfly model object possessing all geometry and therm
            simulation properties.
        _folder_: An optional folder on this computer, into which the THMZ file
            will be written.
        _write: Set to "True" to write the THMZ file from the fairyfly Model.
        run_: Set to "True" to simulate the THMZ file in THERM once it is written.
            This will ensure that results are embedded within the output
            THMZ file such that they can be visualized with the
            "FF Read THERM Result" component.
            _
            This input can also be the integer "2", which will run the whole
            translation and simulation silently (without any batch windows).

    Returns:
        report: A detailed report of the THMZ translation and simulation.
        thmz: The file path of the SQL result file that has been generated on this
            computer. This will be None unless run_ is set to True.
"""

ghenv.Component.Name = 'FF Model to THMZ'
ghenv.Component.NickName = 'ModelToTHMZ'
ghenv.Component.Message = '1.9.1'
ghenv.Component.Category = 'Fairyfly'
ghenv.Component.SubCategory = '1 :: THERM'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

import os

try:
    from fairyfly.config import folders
    from fairyfly.typing import clean_string
    from fairyfly.model import Model
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly:\n\t{}'.format(e))

try:
    from fairyfly_therm.config import folders as therm_folders
    from fairyfly_therm.writer import model_to_thmz
    from fairyfly_therm.run import run_model
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _write:
    # process the simulation folder name
    assert isinstance(_model, Model), \
        'Expected Fairyfly Model for _model input. Got {}.'.format(type(_model))
    model_name = clean_string(_model.display_name)
    _folder_ = os.path.join(folders.default_simulation_folder, model_name) \
        if _folder_ is None else _folder_

    # write the model to THMZ or run it through THERM
    if not run_:
        thmz = os.path.join(_folder_, '{}.thmz'.format(model_name))
        model_to_thmz(_model, thmz)
    else:
        therm_folders.check_therm_version()
        silent = True if run_ > 1 else False
        thmz = run_model(_model, _folder_, silent)



# Fairyfly: A Plugin for Environmental Analysis (GPL)
# This file is part of Fairyfly.
#
# Copyright (c) 2025, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Fairyfly; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create Fairyfly Boundary.
-

    Args:
        _thmz: Path to a THMZ file that has been simulated in THERM. This can be the
            direct output of the "FF Model to THMZ" component as long as
            run_ has been set to True.
        _data_type_: Optional text or an integer to set the type of result data to import from
            the THMZ file. Choose from the following options. (Default: Temperature)
                0 - Temperature (C or F)
                1 - Heat Flux (W/m2 or Btu/h-ft2)
        ip_: Set to True to have all data imported with IP Units (Farenheit and Btu/h-ft2)
            instead of the default SI units of Celsius and W/m2.
        legend_par_: Optional legend parameters from the "LB Legend Parameters"
            that will be used to customize the display of the results.

    Returns:
        report: Reports, errors, warnings, etc.
        shape_geo: A list of breps for the geometry of the Shapes contained within
            the THMZ file. The edges of these shapes can be used to overlay
            wireframe visualizations on top of the result mesh.
        points: The grid of points representing the nodes of the mesh generated by THERM.
        results: A list of numbers that aligns with the points. Each number indicates
            the temeprature (or heat flux) at each of the points.
        mesh: A colored mesh produced from the model's shapes representing the
            temperature or heat flow through the construction.
        legend: A legend showing the values that correspond to the colors of the mesh.
        title: A text object for the study title.
        vis_set: A VisualizationSet object for drawing a detailed version of the
            Therm results in the Rhino scene. This can be connected to
            the "LB Preview Visualization Set" component to display this
            version of the results in Rhino.
"""

ghenv.Component.Name = 'FF Read THERM Result'
ghenv.Component.NickName = 'ThermResult'
ghenv.Component.Message = '1.9.3'
ghenv.Component.Category = 'Fairyfly'
ghenv.Component.SubCategory = '1 :: THERM'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:  # import the core fairyfly dependencies
    from fairyfly_therm.result import THMZResult
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly:\n\t{}'.format(e))

try:  # import the core ladybug dependencies
    from ladybug.graphic import GraphicContainer
    from ladybug.legend import LegendParameters
    from ladybug.color import Colorset
    from ladybug.datatype.temperature import Temperature
    from ladybug.datatype.energyflux import EnergyFlux
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_display.visualization import VisualizationSet
    from ladybug_display.context import ContextGeometry
    from ladybug_display.geometry3d import DisplayLineSegment3D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:  # import the ladybug_rhino dependencies
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.fromgeometry import from_mesh3d, from_point3d, from_face3d
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning, \
        hide_output, show_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

DATA_TYPES = {
    'none': 'temperatures',
    'temperature': 'temperatures',
    '0': 'temperatures',
    'heat flux': 'heat_flux_magnitudes',
    '1': 'heat_flux_magnitudes'
}


if all_required_inputs(ghenv.Component):
    # create the THMZResult object and check to be sure that it has been simulated
    result_obj = THMZResult(_thmz)
    study_mesh = result_obj.mesh
    if study_mesh is None:
        msg = 'No mesh was found within the THMZ file.\nMake sure that the THMZ ' \
            'file has been successfully simulated in THERM.'
        print(msg)
        give_warning(ghenv.Component, msg)

    else:
        # scale the mesh to the current Rhino model units
        conversion = 0.001 / conversion_to_meters()
        study_mesh = study_mesh.scale(conversion)
        shape_geo = result_obj.shape_faces
        shape_geo = [f.scale(conversion) for f in shape_geo]

        # get the results from the THMZ file
        data_attr = DATA_TYPES[str(_data_type_).lower()]
        results = getattr(result_obj, data_attr)
        if results is None:
            msg = 'No steady state results were found within the THMZ file.\n' \
                'Make sure that the THMZ has been set up for steady state simulation ' \
                'and has been successfully simulated in THERM.'
            print(msg)
            give_warning(ghenv.Component, msg)
        else:
            # process the results into the correct units
            if data_attr == 'temperatures':
                units = 'F' if ip_ else 'C'
                study_name = 'Temperature [{}]'.format(units)
                if ip_:
                    results = Temperature().to_ip(results, 'C')[0]
            else:
                units = 'Btu/h-ft2' if ip_ else 'W/m2'
                study_name = 'Heat Flux [{}]'.format(units)
                if ip_:
                    results = EnergyFlux().to_ip(results, 'W/m2')[0]

            # # create the mesh and legend outputs
            l_par = legend_par_ if legend_par_ is not None else LegendParameters()
            if l_par.are_colors_default:
                l_par.colors = Colorset.therm()
            graphic = GraphicContainer(results, study_mesh.min, study_mesh.max, l_par)
            graphic.legend_parameters.title = units
            title = text_objects(
                study_name, graphic.lower_title_location,
                graphic.legend_parameters.text_height * 1.5,
                graphic.legend_parameters.font)

            # create the visual outputs
            study_mesh.colors = graphic.value_colors
            legend = legend_objects(graphic.legend)
            # create the VisualizationSet
            vis_set = VisualizationSet.from_single_analysis_geo(
                'Therm_Results', [study_mesh], results, graphic.legend_parameters)
            wireframe = []
            for face3d in shape_geo:
                for seg in face3d.boundary_segments:
                    wireframe.append(DisplayLineSegment3D(seg, line_width=2))
                if face3d.has_holes:
                    for hole in face3d.hole_segments:
                        for seg in hole:
                            wireframe.append(DisplayLineSegment3D(seg, line_width=2))
            con_outline = ContextGeometry('Wireframe', wireframe)
            vis_set.add_geometry(con_outline)

        # create the visual outputs
        shape_geo = [from_face3d(f) for f in shape_geo]
        points = [from_point3d(pt) for pt in study_mesh.vertices]
        mesh = from_mesh3d(study_mesh)

    # hide the points from the scene
    hide_output(ghenv.Component, 1)
    hide_output(ghenv.Component, 2)



# Fairyfly: A Plugin for Environmental Analysis (GPL)
# This file is part of Fairyfly.
#
# Copyright (c) 2025, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Fairyfly; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create Fairyfly Boundary.
-

    Args:
        _geo: Planar Polyline or Line geometry representing the boundary.
        _name_: Optional text for the name of the boundary. If unspecified,
            a generic one will be automatically assigned.
        _temperature: A number for the temperature at the boundary in degrees Celsius.
            For NFRC conditions, this temperature should be 21C for interior
            boundary conditions and -18 C for winter exterior boundary conditions.
        _film_coeff: A number in W/m2-K that represents the convective resistance of
            the air film at the boundary condition. Typical film coefficient
            values range from 36 W/m2-K (for an exterior condition where
            outdoor wind strips away most convective resistance) to 2.5 W/m2-K
            (for a vertically-oriented interior wood/vinyl surface). For NFRC
            conditions, this should be 26 for exterior boundary conditions and
            around 3 for interior boundary conditions.
        rad_env_: The optional output of the "FF Radiant Environment" component,
            which can be used to customize the radiant properties of the
            boundary (if they differ from the input _temperature and standard
            emissivity of 1 or if there is additional solar heat flux along
            the boundary).
        u_factor_tag_: An optional text string for to define a U-Factor tag along
            the boundary condition. This tag is used tell THERM the boundary on
            which a net U-Value shall be computed. Typical values to input here,
            which are recognizable in LBNL WINDOW include the following.
                * Frame
                * Edge
                * Spacer
                * ShadeInETag
                * ShadeOutETag
                * SHGC Exterior
        rgb_color_: An optional color to set the color of the material when it is
            imported to THERM. All materials from the Fairyfly Therm Library
            already possess colors but materials from the HB-Energy material
            lib will have a randomly-generated color if none is assigned here.

    Returns:
        report: Reports, errors, warnings, etc.
        boundaries: Fairyfly boundaries. These can added to Fairyfly models and
            used in THERM simulation.
"""

ghenv.Component.Name = 'FF Boundary'
ghenv.Component.NickName = 'Boundary'
ghenv.Component.Message = '1.9.1'
ghenv.Component.Category = 'Fairyfly'
ghenv.Component.SubCategory = '0 :: Create'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:  # import the core fairyfly dependencies
    from fairyfly.boundary import Boundary
    from fairyfly.model import Model
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly:\n\t{}'.format(e))

try:  # import the fairyfly-therm extension
    from fairyfly_therm.condition.steadystate import SteadyState
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly_therm:\n\t{}'.format(e))

try:  # import the core ladybug dependencies
    from ladybug.color import Color
    from ladybug_geometry.geometry3d import Polyline3D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:  # import the ladybug_rhino dependencies
    from ladybug_rhino.config import units_system, current_tolerance
    from ladybug_rhino.togeometry import to_polyline3d
    from ladybug_rhino.grasshopper import all_required_inputs, document_counter, \
        longest_list, wrap_output, de_objectify_output, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # check the model tolerance
    tol_msg = Model.check_reasonable_tolerance(units_system(), current_tolerance())
    if tol_msg is not None:
        give_warning(ghenv.Component, tol_msg)

    # create the SteadyState condition object
    name = 'Condition {}'.format(document_counter('condition')) \
        if _name_ is None else _name_
    condition = SteadyState(_temperature, _film_coeff)
    condition.display_name = name
    if rad_env_ is not None:
        emiss, rad_temp, heat_flux = de_objectify_output(rad_env_)
        condition.emissivity = emiss
        condition.radiant_temperature = rad_temp
        condition.heat_flux = heat_flux
    if rgb_color_ is not None:
        condition.color = Color(rgb_color_.R, rgb_color_.G, rgb_color_.B)

    # create the boundary objects
    boundaries = []  # list of boundaries that will be returned
    for j, geo in enumerate(_geo):
        lb_geo = to_polyline3d(geo)
        lb_segs = lb_geo.segments if isinstance(lb_geo, Polyline3D) else [lb_geo]
        ff_boundary = Boundary(lb_segs)
        ff_boundary.display_name = '{} {}'.format(name, j) if len(_geo) > 1 else name
        ff_boundary.properties.therm.condition = condition
        if u_factor_tag_ is not None:
            ff_boundary.properties.therm.u_factor_tag = u_factor_tag_
        boundaries.append(ff_boundary)
    boundaries = wrap_output(boundaries)



# Fairyfly: A Plugin for Environmental Analysis (GPL)
# This file is part of Fairyfly.
#
# Copyright (c) 2025, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Fairyfly; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create Fairyfly Shape.
-

    Args:
        _geo: Rhino Brep or planar Polyline geometry.
        _material: Optional text for the Shape's THERM material to be looked up in the
            material library (see the "FF Search Materials" component for a
            full list of materials that ship with THERM). This input can
            also be a custom SolidMaterial or CavityMaterial object created
            from the corresponding component. Lastly, it can be text for
            the name of a HB-Energy material to be looked up in the material
            library (see the ""HB Search Materials" component for a full
            list of materials that ship with Honeybee).
        rgb_color_: An optional color to set the color of the material when it is
            imported to THERM. All materials from the Fairyfly Therm Library
            already possess colors but materials from the HB-Energy material
            lib will have a randomly-generated color if none is assigned here.

    Returns:
        report: Reports, errors, warnings, etc.
        shape: Fairyfly shapes. These can be added to Fairyfly models and used
            in THERM simulation.
"""

ghenv.Component.Name = 'FF Shape'
ghenv.Component.NickName = 'Shape'
ghenv.Component.Message = '1.9.4'
ghenv.Component.Category = 'Fairyfly'
ghenv.Component.SubCategory = '0 :: Create'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:  # import the core fairyfly dependencies
    from fairyfly.shape import Shape
    from fairyfly.model import Model
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly:\n\t{}'.format(e))

try:  # import the core ladybug dependencies
    from ladybug.color import Color
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:  # import the ladybug_rhino dependencies
    from ladybug_rhino.config import units_system, current_tolerance
    from ladybug_rhino.togeometry import to_face3d
    from ladybug_rhino.grasshopper import all_required_inputs, document_counter, \
        longest_list, wrap_output, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

try:  # import the fairyfly-therm extension
    from fairyfly_therm.lib.materials import solid_material_by_name, \
        cavity_material_by_name
    from fairyfly_therm.material import SolidMaterial, CavityMaterial
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly_therm:\n\t{}'.format(e))

try:  # import the honeybee-energy extension
    from honeybee_energy.lib.materials import opaque_material_by_identifier, \
        window_material_by_identifier
    from honeybee_energy.material.opaque import EnergyMaterial
    from honeybee_energy.material.glazing import EnergyWindowMaterialGlazing
except ImportError as e:
    opaque_material_by_identifier = None
    EnergyMaterial = None
    EnergyWindowMaterialGlazing = None


if all_required_inputs(ghenv.Component):
    # check the model tolerance
    tol_msg = Model.check_reasonable_tolerance(units_system(), current_tolerance())
    if tol_msg is not None:
        give_warning(ghenv.Component, tol_msg)

    #prodess the inputs
    mats = {}  # dictorionary to prevent re-making materials
    shapes = []  # list of shapes that will be returned
    for j, geo in enumerate(_geo):
        mat = longest_list(_material, j)
        if isinstance(mat, str):
            if mat in mats:
                mat = mats[mat]
            else:
                try:
                    mat = solid_material_by_name(mat)
                except ValueError:
                    try:
                        mat = cavity_material_by_name(mat)
                    except ValueError:
                        msg = '"{}" was not found in any of the material ' \
                                'libraries.'.format(mat)
                        try:
                            if opaque_material_by_identifier is None:
                                raise ValueError(msg)
                            mat = opaque_material_by_identifier(mat)
                            mat = SolidMaterial.from_energy_material(mat)
                        except ValueError:
                            try:
                                if window_material_by_identifier is None:
                                    raise ValueError(msg)
                                mat = window_material_by_identifier(mat)
                                mat = SolidMaterial.from_energy_window_material_glazing(mat)
                            except ValueError:
                                raise ValueError(msg)
        else:
            if mat.display_name in mats:
                mat = mats[mat.display_name]
            if EnergyMaterial is not None and isinstance(mat, EnergyMaterial):
                mat = SolidMaterial.from_energy_material(mat)
            if EnergyWindowMaterialGlazing is not None and \
                    isinstance(mat, EnergyWindowMaterialGlazing):
                mat = SolidMaterial.from_energy_window_material_glazing(mat)
            assert isinstance(mat, (SolidMaterial, CavityMaterial)), 'Expected ' \
                'SolidMaterial or CavityMaterial. Got {}.'.format(type(mat))
        if len(rgb_color_) != 0:
            col = longest_list(rgb_color_, j)
            if not mat._locked:
                mat.color = Color(col.R, col.G, col.B)
        mats[mat.display_name] = mat

        lb_faces = to_face3d(geo)
        for lb_face in lb_faces:
            ff_shape = Shape(lb_face)
            ff_shape.display_name = '{} {}'.format(
                mat.display_name, document_counter(mat.display_name))
            ff_shape.properties.therm.material = mat
            shapes.append(ff_shape)
    shapes = wrap_output(shapes)